import pandas as pd
import streamlit as st

# ----- Display Components ----- #


def display_file_headers(headers_dict: dict) -> None:
    """Simply display headers dict."""
    st.header("File Headers", anchor="headers")
    st.write(headers_dict)


def display_file_dataframe(data_frame: pd.DataFrame, height: int, colormap: str) -> None:
    """Simply display dataframe, as a Styler object if a colormap is given."""
    st.header("File Data", anchor="dataframe")
    # Make sure the styler option will allow for the dataframe size
    n_cells = data_frame.shape[0] * data_frame.shape[1]
    pd.set_option("styler.render.max_elements", n_cells)
    styler = (
        data_frame.style.background_gradient(cmap=colormap).highlight_null(color="red")
        if colormap != "None"
        else data_frame
    )
    st.dataframe(styler, height=height)
    st.write(
        "-----\n"
        "Please note that displaying big dataframes is a strain on `Streamlit`'s frontend and will "
        "slow down page refreshes when applying operations. If you loaded a heavy dataframe, consider "
        "un-ticking the `Show File DataFrame` box in the sidebar."
    )



from itertools import zip_longest

import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
import streamlit as st

from tfs_viewer.forms import get_density_plot_params, get_histplot_params, get_scatter_plot_params

# ----- Plotting Functions ----- #


def plotly_line_chart(data_frame: pd.DataFrame) -> None:
    """
    Query user-given options for the plot and craft a plotly Scattergl plot from the data_frame's data.

    Args:
        data_frame (pd.DataFrame): The user loaded TFS data frame.
    """
    versus, plot_quantities, mode, height, errors_x, errors_y = get_scatter_plot_params(data_frame)

    if len(errors_x) not in [0, len(plot_quantities)] or len(errors_y) not in [
        0,
        len(plot_quantities),
    ]:
        st.warning(
            "The amount of properties to plot and of properties to use for error bars do not match. "
            "Some properties will be plotted without error bars."
        )

    fig = go.Figure(layout=go.Layout(height=height))
    for variable, err_x, err_y in zip_longest(plot_quantities, errors_x, errors_y):
        fig.add_trace(
            go.Scattergl(
                x=data_frame[versus].to_numpy(),
                y=data_frame[variable].to_numpy(),
                mode=mode,
                name=variable,
                error_x=(
                    {"type": "data", "array": data_frame[err_x].to_numpy(), "visible": True}
                    if err_x in data_frame.columns
                    else None
                ),
                error_y=(
                    {"type": "data", "array": data_frame[err_y].to_numpy(), "visible": True}
                    if err_y in data_frame.columns
                    else None
                ),
            )
        )
    st.plotly_chart(fig, use_container_width=True)


def plotly_histogram(data_frame: pd.DataFrame) -> None:
    """
    Query user-given options for the plot and craft a plotly histogram plot from the data_frame's data.

    Args:
        data_frame (pd.DataFrame): The user loaded TFS data frame.
    """
    plot_quantities, marginal_mode, histnorm, n_bins, height = get_histplot_params(data_frame)
    if plot_quantities:  # errors if not
        norm_method = None if histnorm == "None" else histnorm
        fig = px.histogram(
            data_frame,
            x=plot_quantities,
            marginal=marginal_mode,
            histnorm=norm_method,
            barmode="overlay",
            height=height,
            nbins=n_bins,
        )
        st.plotly_chart(fig, use_container_width=True)


def plotly_density_contour(data_frame: pd.DataFrame) -> None:
    """
    Query user-given options for the plot and craft a plotly density contour plot from the data_frame's data.

    Args:
        data_frame (pd.DataFrame): The user loaded TFS data frame.
    """
    xcol, ycol, coloring, cmap, reverse_cmap, height = get_density_plot_params(data_frame)
    fig = px.density_contour(data_frame, x=xcol, y=ycol, height=height)
    fig.update_traces(
        contours_coloring=coloring,
        contours_showlabels=True,
        colorscale=None if cmap == "Default" else cmap,
        reversescale=bool(reverse_cmap == "Reversed"),
    )
    st.plotly_chart(fig, use_container_width=True)



from io import StringIO
from tempfile import mkstemp


def handle_file_upload(uploaded) -> tuple:
    """
    Handles a file uploaded with streamlit's file_uploader and writes the contents to a temporary file.
    That temporary file will then be read with tfs-pandas and destroyed. This implementation is convoluted
    but as of now tfs-pandas does not handle buffer objects as returned by streamlit.

    Args:
        uploaded: the uploaded file object given back by streamlit's file_uploader.

    Returns:
        The file object and absolute path of the temporary file where the data was written.
    """
    string_data = StringIO(uploaded.getvalue().decode("utf-8")).read()
    fd, path = mkstemp(suffix=".tfs")
    with open(path, "w") as f:
        f.write(string_data)
    return fd, path



import os

import pandas as pd
import streamlit as st
import tfs

from tfs_viewer.displays import display_file_dataframe, display_file_headers
from tfs_viewer.figures import plotly_density_contour, plotly_histogram, plotly_line_chart
from tfs_viewer.upload import handle_file_upload

GITHUB_BADGE = "https://img.shields.io/badge/GitHub-100000?style=for-the-badge&logo=github&logoColor=white"
GITHUB_URL = "https://github.com/fsoubelet/tfs_viewer_prototype"

# ----- Cached Functions ----- #


@st.cache_data(persist=True)
def load_tfs_file(tfs_file_path: str, index: str, file_obj: int) -> tuple[dict, pd.DataFrame]:
    """
    Loads the chosen TFS file, returns the headers and the dataframe itself. The results are cached for
    efficiency on heavy files. The file will be a temporary file created from uploaded data, and the file
    object is used so we can properly close it.

    Args:
        tfs_file_path (str): absolute path to the file to load.
        index (str): which column to use as inndex during loading.
        file_obj (int): OS-level handle to the opened file as returned by the tempfile.mkstemp function.

    Returns:
        A tuple of the TfsDataFrame's headers (dictionary) and the dataframe itself as a pandas.DataFrame.
    """
    try:
        tfs_df = tfs.read(tfs_file_path, index)
        os.close(file_obj)  # remember to close (and delete) the tempfile
        return tfs_df.headers, pd.DataFrame(tfs_df)
    except Exception as error:  # noqa: BLE001
        st.write(error)


@st.cache_data(persist=True)
def apply_dataframe_query(data_frame: pd.DataFrame, query: str) -> pd.DataFrame:
    return data_frame.query(query)


# ----- Page Config ----- #

st.set_page_config(
    page_title="TFS File Explorer",
    page_icon="🧊",
    layout="wide",
    initial_sidebar_state="expanded",
)
st.title(f"TFS File Explorer [![Github]({GITHUB_BADGE})]({GITHUB_URL})")

# ----- Sidebar Data Manipulation Options ----- #

data_form = st.sidebar.form("Data Options")
data_form.header("Data Manipulation")
chosen_index: str = data_form.text_input("Select Load Index", help="Column to use as index.")
dataframe_query: str = data_form.text_input(
    "Apply Query to Data",
    help="Any query to apply to the dataframe, to be given to `DataFrame.query`. See the [documentation]"
    "(https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DataFrame.query.html) "
    "for usage details.",
)
data_form.form_submit_button("Apply Options")

# ----- Sidebar Data Display Options ----- #

display_form = st.sidebar.form("Display Options")
display_form.header("Display Options")
show_headers: bool = display_form.checkbox(
    "Show File Headers", value=False, help="Whether to display the `header` of the loaded file."
)
show_dataframe: bool = display_form.checkbox(
    "Show File DataFrame", value=True, help="Whether to display the `DataFrame` of the loaded file."
)
dataframe_height: int = display_form.select_slider(
    "DataFrame Display Height", options=list(range(100, 850, 50)), value=400
)
color_map: str = display_form.selectbox(
    "Display Color Map",
    options=["None", "viridis", "plasma", "inferno", "magma", "cividis"],
    help="Which colormap to apply when styling the `DataFrame`.",
)
display_form.form_submit_button("Apply Options")

# ----- Sidebar Visualization Options ----- #

plots_form = st.sidebar.form("Plotting Choices")
plots_form.header("Visualizations")
make_scatterplot: bool = plots_form.checkbox(
    "Craft a ScatterPlot", help="Check this box to create a `Plotly` scatter or line plot."
)
make_histogram: bool = plots_form.checkbox(
    "Craft a Histogram", help="Check this box to create a `Plotly` histogram plot."
)
make_density_plot: bool = plots_form.checkbox(
    "Craft a Density Plot", help="Check this box to create a `Plotly` density plot."
)
plots_form.form_submit_button("Apply")

# ----- Sidebar Data Exploration Options ----- #

# generate_report: bool = st.sidebar.button(
#     "Generate An Exploratory Report",
#     help="Generate and display a `pandas_profiling` report. Beware: This can be time-intensive on big files!",
# )

# ----- Section: User Input ----- #

st.header("Let's Get Your File")
uploaded_file = st.file_uploader("File to load", help="Select your TFS File.")

if uploaded_file is not None:
    file, file_path = handle_file_upload(uploaded_file)
    st.session_state.headers, st.session_state.dataframe = load_tfs_file(file_path, chosen_index, file)

    # Sets desired index if it is changed in the sidebar for an already uploaded file
    if chosen_index != "" and chosen_index in st.session_state.dataframe.columns:
        st.session_state.dataframe = st.session_state.dataframe.set_index(chosen_index)

    dataframe = (
        apply_dataframe_query(st.session_state.dataframe, dataframe_query)
        if dataframe_query != ""
        else st.session_state.dataframe
    )

    # ----- Section: File Data Display ----- #
    if show_headers:
        with st.expander("Headers Section - Click to Fold", expanded=True):
            display_file_headers(st.session_state.headers)
    if show_dataframe:
        with st.expander("Data Section - Click to Fold", expanded=True):
            display_file_dataframe(dataframe, dataframe_height, color_map)

    # ----- Section: Visualizations ----- #
    if make_scatterplot:
        plotly_line_chart(dataframe)
    if make_histogram:
        plotly_histogram(dataframe)
    if make_density_plot:
        plotly_density_contour(dataframe)

# ----- Footer ----- #

st.markdown("-----")

with st.expander("What is This and Who is it For?"):
    st.write(
        "This is a prototype of a simple `streamlit` app to allow my team and I to easily delve into the "
        "contents of a [TFS file](https://mad.web.cern.ch/mad/madx.old/Introduction/tfs.html), which we use "
        "in our work. It uses our own `tfs-pandas` package to handle the format and provides options to "
        "quickly craft simple plots from the data using `streamlit`'s API."
    )
    st.write("**DISCLAIMER**: Functionality may change quickly without notice.")



from __future__ import annotations

from typing import TYPE_CHECKING

import plotly.express as px
import streamlit as st

if TYPE_CHECKING:
    from collections.abc import Sequence

    import pandas as pd


def get_scatter_plot_params(
    data_frame: pd.DataFrame,
) -> tuple[str, Sequence[str], str, int, Sequence[str], Sequence[str]]:
    """
    Form to query the user for scatter plots options with minimal reloading.

    Args:
        data_frame (pd.DataFrame): The user loaded TFS data frame.

    Returns:
        A tuple with the following user defined properties: X-axis column, Y-axis column(s), styling mode for
        the chart, figure height, property(ies) for horizontal error bars and property(ies) for vertical
        error bars.
    """
    scatter_options_form = st.form("Scatter Plot Options")
    scatter_options_form.header("Customize Your Scatter Plot")
    versus, columns, err_x, err_y, height, mode = scatter_options_form.columns(spec=[2, 3, 3, 3, 2, 2])
    with versus:
        versus: str = st.selectbox(
            "Property to Plot Against",
            options=data_frame.columns.to_numpy(),
            help="Select the column that will be on the horizontal axis",
        )
    with columns:
        to_plot: Sequence[str] = st.multiselect(
            "Columns to Plot",
            options=data_frame.columns.to_numpy(),
            help="Select the columns to plot in this chart",
            key="columns_for_scatterplot",
        )
    with height:
        line_chart_height: int = st.select_slider(
            "ScatterPlot Figure Height", options=list(range(200, 1500, 50)), value=700
        )
    with err_x:
        horizontal_errors: Sequence[str] = st.multiselect(
            "Hozirontal Error Bars",
            options=["", *data_frame.columns.tolist()],
            help="Property to use for horizontal error bars. The first property provided here will be "
            "used for errors of the first property provided there and so on. Be aware that a mismatch "
            "in the number of inputs to plot and to use as error bars means some properties will be "
            "plotted with no error bars.",
        )
    with err_y:
        vertical_errors: Sequence[str] = st.multiselect(
            "Vertical Error Bars",
            options=["", *data_frame.columns.tolist()],
            help="Property to use for vertical error bars. The first property provided here will be "
            "used for errors of the first property provided there and so on. Be aware that a mismatch "
            "in the number of inputs to plot and to use as error bars means some properties will be "
            "plotted with no error bars.",
        )
    with mode:
        mode: str = st.selectbox(
            "Styling of the line chart",
            options=["lines", "markers", "lines+markers"],
            help="The styling of the scatter plot data",
        )
    scatter_options_form.form_submit_button("Submit and Update Plot")
    return versus, to_plot, mode, line_chart_height, horizontal_errors, vertical_errors


def get_histplot_params(data_frame: pd.DataFrame) -> tuple[Sequence[str], str, str, int, int]:
    """
    Form to query the user for histogram plot options with minimal reloading.

    Args:
        data_frame (pd.DataFrame): The user loaded TFS data frame.

    Returns:
        A tuple with the following user defined properties: columns to plot, styling mode for the
        distribution upper plot, normalization routine for the histogram bins, number of bins and figure
        height.
    """
    histplot_options_form = st.form("Histogram Plot Options")
    histplot_options_form.header("Customize Your Histogram Plot")
    columns, marginal_mode, normalization_mode, height, n_bins = histplot_options_form.columns(spec=[4, 3, 3, 3, 2])
    with columns:
        to_plot: Sequence[str] = st.multiselect(
            "Columns to Plot",
            options=data_frame.columns.to_numpy(),
            help="Select the columns to plot in this chart",
            key="columns_for_histogram",
        )
    with marginal_mode:
        mode: str = st.selectbox(
            "Styling of distribution plot",
            options=["box", "violin", "rug"],
            help="The type of distribution representation used for the upper axis",
        )
    with normalization_mode:
        histnorm: str = st.selectbox(
            "Normalization Routine",
            options=["None", "percent", "probability", "density", "probability density"],
            help="Bin normalization method. If None is selected, then the simple value counts are used",
        )
    with height:
        histogram_plot_height: int = st.select_slider(
            "Histogram Figure Height", options=list(range(200, 1500, 50)), value=700
        )
    with n_bins:
        nbins: int | float = st.number_input(  # careful streamlit might infer float from int inputs
            "Number of Bins",
            value=100,
            step=25,
            min_value=5,
            max_value=1000,
            help="Number of bins in the histogram",
        )
    histplot_options_form.form_submit_button("Submit and Update Plot")
    return to_plot, mode, histnorm, int(nbins), histogram_plot_height


def get_density_plot_params(data_frame: pd.DataFrame) -> tuple[str, str, str, str, bool, int]:
    """
    Form to query the user for density plot options with minimal reloading.

    Args:
        data_frame (pd.DataFrame): The user loaded TFS data frame.

    Returns:
        A tuple with the following user defined properties: X-axis column, Y-axis column, coutour type,
        color map to use, whether to reverse the colormap and figure height.
    """
    density_plot_options_form = st.form("Density Plot Options")
    density_plot_options_form.header("Customize Your Density Plot")
    xaxis, yaxis, contours, cmap, height, cmap_reverse = density_plot_options_form.columns(spec=[3, 3, 3, 3, 3, 2])
    with xaxis:
        xaxis_var: str = st.selectbox(
            "Property on the Horizontal Axis",
            options=data_frame.columns.to_numpy(),
            help="Select the column that will be on the horizontal axis",
        )
    with yaxis:
        yaxis_var: str = st.selectbox(
            "Property on the Vertical Axis",
            options=data_frame.columns.to_numpy(),
            help="Select the column that will be on the vertical axis",
        )
    with contours:
        coloring: str = st.selectbox(
            "Contour Types",
            options=["fill", "heatmap", "lines", "none"],
            help="Whether to fill the contours by extrapolating data",
        )
    with cmap:
        colorscale: str = st.selectbox(
            "Color Map",
            options=["Default"] + [cmap.capitalize() for cmap in px.colors.named_colorscales()],
            help="Which sequencial colormap to use for this plot",
        )
    with height:
        density_plot_height: int = st.select_slider(
            "Histogram Figure Height", options=list(range(200, 1500, 50)), value=700
        )
    with cmap_reverse:
        reverse_cmap: str = st.selectbox(
            "Colormap Scale",
            options=["Classic", "Reversed"],
            help="Whether to reverse the colormap for the plot",
        )
    density_plot_options_form.form_submit_button("Submit and Update Plot")
    return xaxis_var, yaxis_var, coloring, colorscale, reverse_cmap, density_plot_height
